package experiment;

import experiment.entity.Link;
import experiment.entity.Node;

import java.io.Serializable;
import java.util.Collection;
import java.util.HashMap;
import visual.Crawler;
import walker.ExtendedXMLStreamReader;
import walker.StaxParser;
import walker.Weyes;
import walker.exceptions.badSyntaxException;

import javax.vecmath.Vector3d;

/**
 * Created by IntelliJ IDEA.
 * User: andy
 * Date: Mar 22, 2010
 *
 * Contains all the data that corresponds to a single experiment state. Implements Serializable in order to allow
 * easy storing.
 */
public class ExperimentState implements Serializable {

    /**
     * The state's current time. It matches the time value of the latest update. The value is relative to the 
     * start time of the experiment.
     */
    private double timestamp;

    /**
     * Reference of all Node objects in association with their ids
     */
    private int nodeNumber;
    private HashMap<String,Node> nodeMap;
    private int linkNumber;
    private transient ExtendedXMLStreamReader reader;
    private long byteOffset;

    public ExperimentState() {
        this.timestamp = 0;
        this.nodeNumber = 0;
        this.nodeMap = new HashMap<String, Node>();
        this.linkNumber = 0;
        this.byteOffset = -1;
    }

    /**
     * Adds a Node to the current state.
     *
     * @param node the node object to be added.
     */
    public synchronized void add(Node node){
        nodeMap.put(node.getId(), node);
        nodeNumber++;
        // in case of a node discovered during trace parsing,
        if (StaxParser.initialized){
            Weyes.crawler.engine3d.addEntity(node);
        }
    }

    /**
     * Adds a Link to the current state.
     *
     * @param link the link object to be added.
     */
    public synchronized void add(Link link){
        linkNumber++;
        // in case of a link discovered during trace parsing,
        if (StaxParser.initialized){
            Weyes.crawler.engine3d.addEntity(link);
        }
    }

    /**
     * @return the number of nodes currently in the experiment
     */
    public synchronized int getNodeNumber(){
        return nodeNumber;
    }

    /**
     * @return the current number of Links in the experiment
     */
    public synchronized int getLinkNumber() {
        return linkNumber;
    }

    /**
     * Indexes a node based on its id.
     *
     * @param nodeId the node's id
     * @return a reference to the Node object
     */
    public synchronized Node indexNode(String nodeId){
        if (nodeId == null){
            return null;
        }
        
		return nodeMap.get(nodeId);
    }

    /**
     * Indexes a link.
     *
     * @param sourceId the Link's source Node id.
     * @param targetId the Link's target Node id.
     * @return the Link or null.
     */
    public synchronized Link indexLink(String sourceId, String targetId){
        if (sourceId != null && targetId != null){
            return indexNode(sourceId).indexLink(targetId);
        }
        return null;
    }

    /**
     * @return the state's time.
     */
    public synchronized double getTime() {
        return timestamp;
    }

    /**
     * Is called to change a state's timestamp. It only changes the timestamp if the new time value
     * is greater than the old one.
     *
     * @param timestamp is a long variable containing the new timestamp.
     * @throws badSyntaxException if the timestamp is smaller than the state's timestamp.
     */
    public synchronized void setTime(double timestamp) throws badSyntaxException {
        if (this.timestamp <= timestamp){
            this.timestamp = timestamp;
        } else {
            throw new badSyntaxException("Invalid timestamp value");
        }
    }

    /**
     * @return a collection of all Nodes in this state.
     */
    public synchronized Collection<Node> getNodeSet(){
        return nodeMap.values();
    }

    /**
     * Each ExperimentState object is bound to one ExtendedXMLStreamReader. When the object is cloned or
     * serialized and stored, the ExtendedXMLStreamReader is also cloned/stored. This way the parsing may continue
     * from any cloned/serialized state.
     *
     * @return the ExtendedXMLStreamReader that updates this state.
     */
    public ExtendedXMLStreamReader getXMLReader() {
        return reader;
    }

    public void setXMLReader(ExtendedXMLStreamReader reader) {
        this.reader = reader;
    }

    /**
     * Tries to find the Node that corresponds to a position in 3d space. Is called when a user picks an object.
     *
     * @param position a three dimensional Vector containing the x, y, z double values.
     * @return The indexed Node or null;
     */
    public synchronized Node indexNode(Vector3d position) {
        Collection<Node> nodes = nodeMap.values();
        for (Node node : nodes){
            if (node.getPosition().same(position)){
                return node;
            }
        }
        return null;
    }

    /**
     * As a state contains parsed experiment data it is also associated with a byte offset, which points to the
     * location of the file where the parsing has reached.
     *
     * @return the long value of the experiment state's byte offset.
     */
    public long getByteOffset() {
        return byteOffset;
    }

    /**
     * As a state contains parsed experiment data it is also associated with a byte offset which points to the
     * location of the file where the parsing has reached. The byte offset is set when the state is serialized.
     * To compensate for a previous jump, StaxParxer's jumpOffset variable is also added.
     */
    public void setByteOffset() {
        this.byteOffset = Crawler.getReader().getReader().getLocation().getCharacterOffset() + StaxParser.jumpOffset;
    }

}
