package experiment;

import experiment.event.Event;

import java.util.concurrent.LinkedBlockingQueue;
import walker.exceptions.badSyntaxException;

/**
 * Created by IntelliJ IDEA.
 * User: andy
 * Date: Mar 11, 2010
 *
 * Based on the real set of data a number of scenarios can be specified. For example, a failure can be
 * simulated by disabling a node or modifying the rssi value on a link at a certain moment in time.
 * As the real data will be used in simulators, WiseML provides a section in the description in which
 * these scenarios can be specified.
 */
public class Scenario {

    // the scenario id
    private String id;
    // a queue that stores all scenario events
    private LinkedBlockingQueue<Event> eventQueue;

    /**
     * Default Constructor
     * @param id the scenario's id
     */
    public Scenario(String id) {
        this.id = id;
        eventQueue = new LinkedBlockingQueue<Event>();
    }


    /**
     * Applies the youngest event after checking if it's time.
     * todo: change the visualizationWindow name and concept.
     *
     * @param visualizationWindow the time window in milliseconds before the next update
     * @param state defines which of the experiment states to apply it to.
     * @return true if the youngest event was applied
     * @throws walker.exceptions.badSyntaxException in case of a malformed event.
     */
    public boolean apply(double visualizationWindow, ExperimentState state) throws badSyntaxException {
        if (isTime(state, visualizationWindow)){
            pullEvent().apply(state);
            return true;
        }
		return false;
    }

    /**
     * Tests if it's the right time to apply the younger event.
     *
     * @param state defines which of the experiment states to check.
     * @param visualizationWindow the time window in milliseconds before the next update.
     * @return true if it is time to apply the youngest event of this scenario.
     */
    public boolean isTime(ExperimentState state, double visualizationWindow){
        double tailEventTime = eventQueue.peek().getTimestamp();
        return  (tailEventTime > state.getTime() && tailEventTime <= state.getTime() - visualizationWindow);
    }

    /**
     * Add an ecent to the scenario's queue
     * @param ev the event to be added
     */
    public void addEvent(Event ev){
        eventQueue.add(ev);
    }

    /**
     * Returns the younger event
     * @return the tail of the event queue
     */
    public Event pullEvent(){
        return eventQueue.poll();
    }

    /**
     * @return the Scenario's unique id
     */
    public String getId() {
        return id;
    }
}
