package experiment.entity;

import experiment.property.Position;
import experiment.property.Capability;
import experiment.ExperimentState;

import java.util.HashMap;
import java.util.Collection;
import walker.utilities.SerialClone;

/**
 * Created by IntelliJ IDEA.
 * User: andy
 * Date: Mar 10, 2010
 *
 * Represents an experiment Node and stores all the node's data
 */
public class Node extends Entity {

    //  unused? private static int nodeNumber = 0;

    /**
     * Static default variables
     */
    public static Position defaultPosition = new Position(0,0,0,0,0);
    public static boolean defaultGateway = false;
    public static String defaultProgDetails = "";
    public static String defaultNodeType = "";
    public static String defaultDescription = "";
    // Stores default capabilities in a HashMap with the capability name as key.
    private static HashMap<String, Capability> defaultCapabilityMap = new HashMap<String, Capability>();

    /**
     * Node variables
     */
    // the Node3d's universal String id that is parsed from the wiseml file.
    private String id;
    
    private Position position;
    private boolean gateway;
    private String programDetails;
    private String nodeType;
    private String description;

    // Stores all links that have this Node3d as a source
    private HashMap<String, Link> links = new HashMap<String, Link>();

    /**
     * Full Constructor. Put null to all not available and not required attributes.
     *
     * @param state the object that contains all the data for a given experiment state
     * @param id the node's id
     * @param position the node's position
     * @param gateway it is a Boolean value in a String indicating if the node is a gateway or not. User can specify a default value.
     * @param programDetails contains textual information describing the software image loaded on the node.
     * @param nodeType contains textual information about the type of the device.
     * @param description this is a textual description of the node.
     */
    public Node(ExperimentState state, String id, Position position, String gateway, String programDetails, String nodeType, String description) {
        // Set the Node3d's state
        super(state);

        this.id = id;

        initialize();

        if (position!=null){
            this.position = position;
        }
        if (gateway!=null){
            this.gateway = Boolean.parseBoolean(gateway);
        }
        if (programDetails!=null){
            this.programDetails = programDetails;
        }
        if (nodeType!=null){
            this.nodeType = nodeType;
        }
        if (description!=null){
            this.description = description;
        }

        // Initialize the capability map with the default values
        capabilityMap = SerialClone.clone(defaultCapabilityMap);

        isNode = true;

        // add the node to its state.
        getState().add(this);

        updated();

    }

    /**
     * Set's all Node variables to the default value
     */
    public void initialize(){
        this.position = defaultPosition;
        this.gateway = defaultGateway;
        this.programDetails = defaultProgDetails;
        this.nodeType = defaultNodeType;
        this.description = defaultDescription;
    }

    
    /*
     *
     * Static Methods
     *
     */

    /**
     * Is called during experiment configurarion to set the default node capabilities.
     *
     * @param dC the capability to be set
     */
    public static void addDefaultCapability(Capability dC){
        if (!defaultCapabilityMap.containsKey(dC.getName())){
            defaultCapabilityMap.put(dC.getName(), dC);
        }
    }


    /*
     *
     * Non-Static Methods
     *
     */

    /**
     * Indexes the links that have this node as source.
     * @param targetId the id of the node that is the target of the desired link
     * @return a Link with this Node as source and the targetId as target
     */
    public synchronized Link indexLink(String targetId){
        return links.get(targetId);
    }

    /**
     * Adds a link to this Node's linkMap.
     *
     * @param link the link has this node as source
     * @param targetId the id of the link's target Node
     */
    public synchronized void addLink(Link link, String targetId){
        links.put(targetId, link);
    }

    /**
     * @return a collection of all the Links that have this Node as source.
     */
    public synchronized Collection<Link> getLinkSet(){
        return links.values();
    }

    /**
     * Set the Node as a gateway.
     *
     * @param gateway true if this is a gateway.
     */
    public synchronized void setGateway(boolean gateway) {
        this.gateway = gateway;
        updated();
    }

    /**
     * Set the NodeType.
     *
     * @param nodeType a String containing the new node type
     */
    public synchronized void setNodeType(String nodeType) {
        this.nodeType = nodeType;
        updated();
    }


    /**
     * Update the Node's position.
     *
     * @param position the new position
     */
    public synchronized void updatePosition(Position position) {
        this.position = position;
        updated();
    }

    /**
     * @return the node's current position.
     */
    public synchronized Position getPosition() {
        return position;
    }

    /**
     * This method is not synchronized as the program Details are not expected to change after initialization.
     *
     * @return A description of the software image loaded on the node.
     */
    public String getProgramDetails() {
        return programDetails;
    }

    /**
     * This method is not synchronized as the description is not expected to change after initialization.
     *
     * @return a description of the Node.
     */
    public String getDescription() {
        return description;
    }

    /**
     * This method is not synchronized as the node id cannot change.
     *
     * @return the Node's unique id.
     */
    public String getId() {
        return id;
    }

    /**
     * @return true if this Node is a gateway.
     */
    public synchronized boolean isGateway() {
        return gateway;
    }

    /**
     * This method is not synchronized as the Node type is not expected to change after initialization.
     *
     * @return a description of the device.
     */
    public String getNodeType() {
        return nodeType;
    }

}
