package experiment.event.capability;

import visual.entity.Link3d;
import walker.Weyes;
import walker.exceptions.badSyntaxException;
import experiment.ExperimentState;
import experiment.entity.Link;
import experiment.entity.Node;
import experiment.property.Rssi;

/**
 * Created by IntelliJ IDEA.
 * User: andy
 * Date: Mar 26, 2010
 *
 * Includes the data update information for a single Link in a single time.
 */
public class LinkValueEvent extends ValueEvent {

    String source = null, target = null;
    Rssi rssi = null;

    /**
     * Default Constructor. Takes only the link's source and target node ids. If a capability has to be set,
     * use the relative method after constructing the Event.
     *
     * @param timestamp the exact experiment moment that this event occured.
     * @param source the source node's id.
     * @param target the target node's id.
     * @param rssi the new Rssi value. Can be null.
     */
    public LinkValueEvent(double timestamp, String source, String target, Rssi rssi) {
        this.timestamp = timestamp;
        this.source = source;
        this.target = target;
        this.rssi = rssi;
    }

    /**
     * Applies this event to the experiment.
     *
     * @param state the experiment state where this event belong to.
     * @return true if event was successfully applied.
     */
    public boolean apply(ExperimentState state) throws badSyntaxException {
        // check that this is a valid linkValueEvent.
        if (source == null || target == null){
            return false;
        }

        // Try to index the Node.
        Node sourceNode = state.indexNode(source);
        Node targetNode = state.indexNode(target);

        /*
         * If the an Entity doesn't exist, add it to the experiment setup.
         */
        if (sourceNode == null){
            sourceNode = new Node(state, source, null, null, null, null, null);
            state.add(sourceNode);
        }

        if (targetNode == null){
            targetNode = new Node(state, target, null, null, null, null, null);
            state.add(targetNode);
        }

        // Try to index the Link.
        Link l = sourceNode.indexLink(target);
        if (l == null){
            l = new Link(state, sourceNode, targetNode, null, null, null);
        }

        // Update the corresponding states timestamp
        state.setTime(this.timestamp);

        // Update its rssi value
        if (this.rssi!=null){
            l.setRssi(rssi);
        }

        // loop through the directiveList and update all known capabilities
        for (Directive aDirective : directiveList) {
            l.setCapability(aDirective.getKey(), aDirective.getData());

            if (Weyes.LinkDisplayCapability!=null && aDirective.getKey()!=null && aDirective.getKey().equals(Weyes.LinkDisplayCapability) && aDirective.getData()!=null){
                if (l.getEntity3d()!=null){
                    ((Link3d)l.getEntity3d()).changeColor(aDirective.getData());
                }
            }
        }

        return true;
    }
}
