package visual.entity;

import experiment.property.Position;
import experiment.entity.*;
import experiment.entity.Node;
import experiment.entity.Link;
import experiment.ExperimentState;

import javax.media.j3d.*;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;
import java.util.HashSet;

import walker.Weyes;

/**
 * Created by IntelliJ IDEA.
 * User: andy
 * Date: Apr 8, 2010
 *
 * An abstract class for creating and displaying experiment.Entity objects in the 3d canvas.
 * Every Entity object (Node/Link) that is or is going to be displayed, is bound to an Entity3d object and visa versa.
 */
public abstract class Entity3d {

    // Each Entity3d is placed under a unique Transform Group in the scene Graph.
    protected TransformGroup TG;
    // Each experiment Entity has a 1-1 relation with an entity3d.
    protected Entity entity;
    // Is going to be calculated as to normalize entity positioning so that everything fits the scene's bounding box.
    protected static double factor = 1;

    // Stores all entity3d instances. It is used for rewiring after a seek.
    private static HashSet<Entity3d> entity3dSet = new HashSet<Entity3d>();

    /**
     * Defines allowed scale operations.
     */
    public static enum SCALE {
        ENLARGE, SHRINK
    }

    /**
     * Default Constructor. Associates a 3d representation with its corresponding Entity.
     *
     * @param entity the corresponding experiment entity.
     */
    public Entity3d(Entity entity){
        entity.setEntity3d(this);
        this.entity = entity;
        entity3dSet.add(this);
    }

    /**
     * Is called during a seeking operation or after a reset operation to update and rewire each
     * Entity, Entity3d correlation.
     *
     * @param state the new ExperimentState containing all updated Entity instances.
     */
    public static void rewire(ExperimentState state){

        for (Entity3d iter : entity3dSet){
            if (iter.getEntity().isNode()){
                Node node = state.indexNode(((Node)iter.getEntity()).getId());
                if (node != null){
                    // rewire
                    iter.entity = node;
                    node.entity3d = iter;
                    iter.show();

                // there is no node yet
                } else {
                     iter.hide();
                }
            } else {
                Link oldLink = (Link) iter.getEntity();
                Link newlink = state.indexLink(oldLink.getSource().getId(),oldLink.getTarget().getId());
                if (newlink != null){
                    // rewire
                    iter.entity = newlink;
                    newlink.entity3d = iter;
                    iter.show();
                    // update the link's color
                    if (Weyes.LinkDisplayCapability != null){
                        String value = newlink.getCapability(Weyes.LinkDisplayCapability).getValue();
                        if (value != null){
                            ((Link3d)iter).changeColor(value);
                        }
                    }

                // there is no link yet
                } else {
                    iter.hide();
                }
            }
        }
    }

    /**
     * As soon an Entity3d is constructed, it is saved to the entity3dSet. Even if it may not be seen, or if it is not
     * associated with an Entity it still exists. This usually occurs when the user seeks backward. When a new Entity is
     * constructed with the same id, the Entity3d is rewired and re-seen again.
     *
     * @param entity holds the experiment information
     * @return the 3d information associated with the experiment information.
     */
    public static Entity3d get(Entity entity){
        if (entity.isNode()){
            for (Entity3d iter : entity3dSet){
                Entity tmp = iter.getEntity();
                if (tmp.isNode() && ((Node) tmp).getId().equals(((Node) entity).getId())){
                    return iter;
                }
            }
        } else {
           for (Entity3d iter : entity3dSet){
                Entity tmp = iter.getEntity();
                if (!tmp.isNode() && ((Link) tmp).getSource().getId().equals(((Link) entity).getSource().getId()) && ((Link) tmp).getTarget().getId().equals(((Link) entity).getTarget().getId())){
                    return iter;
                }
            }
        }
        return null;
    }

    /**
     * If an entity has been constructed during experiment visualization, then in case of seeking or stop, when the
     * rewire() method is called, it may be that even though the entity3d is displayed, it has not yet been constracted.
     * In this case the entity3d is hidden until the entity is reconstructed.
     */
    public abstract void hide();

    /**
     * If during experiment visualization a new Entity is added and the corresponding Entity3d is hidden,
     * show the Entity3d.
     */
    public abstract void show();
    /**
     * @return the tranform Group containing the 3d entity's positioning information.
     */
    public TransformGroup getTG() {
        return TG;
    }

    /**
     * @return the experiment.Entity object that is associated with this 3d entity.
     */
    public Entity getEntity(){
        return entity;
    }

    /**
     * Define the appearance of the 3d Entity.
     *
     * @return an Appearance object.
     */
    public abstract Appearance createAppearance();

    /**
     * Move the 3d Entity to a new Position.
     *
     * @param p is the new experiment.entity.Position where the 3d Entity is going to be moved.
     */
    public abstract void move(Position p);

    /**
     * Calculate the normalization factor. The factor is calculated so that the furthest node is positioned
     * at the edge of the plane.
     *
     * @param state the experiment state conatining all node data.
     */
    public static void setFactor(ExperimentState state){
        double distance = 0;
        for (experiment.entity.Node node :state.getNodeSet()){
            Point3d tmp = new Point3d(node.getPosition().getVector3d());
            if (tmp.distance(new Point3d(0,0,0)) > distance){
                distance = tmp.distance(new Point3d(0,0,0));
            }
        }

        factor = distance/25; //todo make 25 a final variable associated with grid.
    }

    /**
     * Normalize entity positions to achieve better view of the experiment.
     *
     * @param vector3d the actual position of the entity.
     * @return the normalized position.
     */
    protected Vector3d normalize(Vector3d vector3d) {
        return new Vector3d(vector3d.getX()/factor, vector3d.getY()/factor, vector3d.getZ()/factor);
    }

    /*
     * Animation methods
     */

    /**
     * Either enlarges or shrinks the 3d entity. It is used with picking and is called by FrontEnd when the monitored
     * entity changes.
     *
     * @param mode enlarge or shrink?
     */
    public abstract void scale(SCALE mode);

    /**
     * Makes the 3d entity object flash red.
     */
    public abstract void reden();

    /**
     * Makes a 3d Entity appear as enabled or disabled.
     */
    public abstract void updateStatus();


}
