package visual.entity;

import com.sun.j3d.utils.geometry.Cylinder;
import experiment.entity.Link;
import experiment.property.Position;
import static visual.Palette.darkblue;
import visual.Palette;

import javax.media.j3d.*;
import javax.vecmath.Color3f;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

/**
 * Created by IntelliJ IDEA.
 * User: andy
 * Date: Apr 6, 2010
 *
 * Creates a 3d line that links 2 experiment Nodes. It is the visual 3d representation of a Link.
 * It also contains the line's transform group.
 */
public class Link3d extends Entity3d {

    private Shape3D line3d;

    /**
     * Default Constructor. Creates the line that connects two nodes.
     *
     * @param link the corresponding experiment.Link.
     */
    public Link3d(Link link) {
        // call Entity3d constructor
        super(link);

        // define the Positioning transfom and the Transform Group that will hold the line.
        Transform3D translate = new Transform3D();
        TG = new TransformGroup(translate);
        TG.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
        TG.setCapability(TransformGroup.ALLOW_CHILDREN_WRITE);

        // create the line that link's two Nodes
        LineArray line = new LineArray(2, LineArray.COORDINATES);
        line.setCoordinate( 0, new Point3d(normalize(link.getSource().getPosition().getVector3d())));
        line.setCoordinate( 1, new Point3d(normalize(link.getTarget().getPosition().getVector3d())));
        Appearance lineAppearance = createAppearance();
        line3d = new Shape3D(line, lineAppearance);
        line3d.setCapability(Shape3D.ALLOW_APPEARANCE_WRITE);

        // Associate the line3d with the Link
        line3d.setUserData(link.getSource().getId() + "-" + link.getTarget().getId());
        line3d.setName("link");

        TG.addChild(line3d);

        // allow the Line to be picked.
        line3d.setCapability(javax.media.j3d.Node.ENABLE_PICK_REPORTING);
    }

    /**
     * If an entity has been constructed during experiment visualization, then in case of seeking or stop, when the
     * rewire() method is called, it may be that even though the entity3d is displayed, it has not yet been constracted.
     * In this case the entity3d is hidden until the entity is reconstructed.
     */
    public void hide() {
        line3d.getAppearance().getTransparencyAttributes().setTransparency(1);
    }

    /**
     * If during experiment visualization a new Entity is added and the corresponding Entity3d is hidden,
     * show the Entity3d.
     */
    public void show() {
        line3d.getAppearance().getTransparencyAttributes().setTransparency(0);
        resetColor();
    }

    /**
     * Define the appearance of the 3d Link.
     *
     * @return an Appearance object.
     */
    public Appearance createAppearance() {

        Appearance app = new Appearance();

        TransparencyAttributes trans = new TransparencyAttributes();
        trans.setCapability(TransparencyAttributes.ALLOW_VALUE_WRITE);
        app.setTransparencyAttributes(trans);
        trans.setTransparency(0);
        app.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_WRITE);

        ColoringAttributes ca = new ColoringAttributes(darkblue, ColoringAttributes.FASTEST);
        app.setColoringAttributes(ca);
        app.setCapability(Appearance.ALLOW_COLORING_ATTRIBUTES_WRITE);
        ca.setCapability(ColoringAttributes.ALLOW_COLOR_WRITE);

        LineAttributes lineAttributes = new LineAttributes(0.5f, LineAttributes.PATTERN_SOLID, true);
        app.setCapability(Appearance.ALLOW_LINE_ATTRIBUTES_WRITE);
        lineAttributes.setCapability(LineAttributes.ALLOW_PATTERN_WRITE);
        lineAttributes.setCapability(LineAttributes.ALLOW_WIDTH_WRITE);

        app.setLineAttributes(lineAttributes);
        return app;
    }

    /**
     * Move the 3d Entity to a new Position.
     *
     * @param p is the new experiment.entity.Position where the 3d Entity is going to be moved.
     */
    public void move(Position p) {      //todo if a node is moved then its links should also be moved.

    }

    /**
     * Either enlarges or shrinks the 3d link. It is used with picking and is called by FrontEnd when the monitored
     * entity changes.
     *
     * @param mode enlarge or shrink?
     */
    public void scale(SCALE mode) {

        LineAttributes line = line3d.getAppearance().getLineAttributes();
        if (mode == SCALE.ENLARGE){
            line.setLineWidth(4f);
            line3d.getAppearance().getTransparencyAttributes().setTransparency(0.4f);

        } else if (mode == SCALE.SHRINK){
            line.setLineWidth(0.5f);
            line3d.getAppearance().getTransparencyAttributes().setTransparency(0);

        }

        line3d.getAppearance().setLineAttributes(line);
    }

    /**
     * Makes the 3d entity object flash red.
     */
    public void reden() {

    }

    /**
     * Changes the color of this line object.
     *
     * @param value the LinkDisplayCapability.
     */
    public void changeColor(String value){
        double v;
        try {
            v = Double.parseDouble(value);
        } catch (Exception e) {
            System.out.println("Malformed LinkDisplayCapability value");
            return;
        }

        Color3f color;

        if (v <= 0){
            color = Palette.whitegray;
        } else if (v <= 10){
            color = Palette.r1;
        } else if (v <= 20){
            color = Palette.r2;
        } else if (v <= 30){
            color = Palette.r3;
        } else if (v <= 40){
            color = Palette.r4;
        } else if (v <= 50){
            color = Palette.r5;
        } else if (v <= 60){
            color = Palette.r6;
        } else if (v <= 70){
            color = Palette.r7;
        } else if (v <= 80){
            color = Palette.r8;
        } else if (v <= 90){
            color = Palette.r9;
        } else if (v <= 100){
            color = Palette.r10;
        } else {
            color = Palette.r11;
        }

        line3d.getAppearance().getColoringAttributes().setColor(color);
    }

    /**
     * Resets the color of this line3d to darkblue.
     */
    public void resetColor(){
       line3d.getAppearance().getColoringAttributes().setColor(darkblue);
    }

    /**
     * Makes a 3d Entity appear as enabled or disabled. If the Link is disabled, the line appears dashed.
     * NOT TESTED!
     */
    public void updateStatus() {
        if (entity.isEnabled()){
           line3d.getAppearance().getLineAttributes().setLinePattern(LineAttributes.PATTERN_SOLID);
        } else {
           line3d.getAppearance().getLineAttributes().setLinePattern(LineAttributes.PATTERN_DASH);
        }
    }

    /**
     * Overrides the default updateStatus().
     * It is used to make the links of a disabled node also appear disabled.
     * NOT TESTED!
     *
     * @param state false to make this link3d appear disabled. True does nothing.
     */
    public void updateStatus(boolean state){
        if (!state){
            line3d.getAppearance().getLineAttributes().setLinePattern(LineAttributes.PATTERN_DASH);
        }
    }

    /**
     * Draw a cylinder instead of a line.
     * Not in use. Rotations don't work.
     *
     * @param source the source Node position.
     * @param target the target Node position.
     * @param translate the used Transform3D.
     * @return the Cylinder Primitive.
     */
    public Cylinder createCylinder(Vector3d source, Vector3d target, Transform3D translate){

        // Find the center
        Vector3d center = new Vector3d(source);
        center.add(target);
        center.set(center.getX()/2, center.getY()/2, center.getZ()/2);

        // Find the length of the cylinder.
        double dx = (target.getX()-source.getX());
        double dy = (target.getY()-source.getY());
        double dz = (target.getZ()-source.getZ());
        float length = (float) Math.sqrt(dx*dx + dy*dy + dz*dz);

        // Find the rotation on Z-axis.
        double rotZ=0;
        if (dx == 0){
            rotZ=0;
        }else if (dy == 0){
            //CHANGE THIS...
            rotZ=Math.PI/2; //if this is negative, remove the -rotX below
            if (dx < 0){
                rotZ = -rotZ;
            }
        }else {
            rotZ=-Math.atan(Math.abs(dx)/Math.abs(dy));
            if (dy < 0){
                if (dx>0){
                    rotZ = Math.PI+rotZ;
                }else{
                    rotZ = Math.PI*3-rotZ;
                }
            }else{
                if (dx>0){
                    rotZ=-rotZ;
                }
            }
        }
        rotZ=-rotZ;

        // Find the rotation on the X axis.
        double rotX = Math.atan ( Math.abs(target.getZ()-center.getZ()) / (0.5*Math.sqrt(dx*dx+dy*dy)) );
        rotX=-rotX;//If this goes, alter the CHANGE THIS above
        if (target.getZ()-center.getZ() > 0){
            //CHANGE THIS...
            rotX = -rotX;
        }

        Transform3D rotationZ = new Transform3D();
        Transform3D rotationX = new Transform3D();
        rotationZ.rotZ(rotZ);
        rotationX.rotX(rotX);

        translate.mul(rotationZ);
        translate.mul(rotationX);

        return new Cylinder(0.02f, length, 1, 15, 2, createAppearance());
    }
}
