package walker;

import experiment.Setup;
import experiment.property.Capability;
import experiment.property.Position;
import experiment.property.Rssi;
import walker.exceptions.badSyntaxException;
import walker.exceptions.endDocumentException;
import static walker.utilities.Output.print;

import javax.xml.stream.*;
import java.io.InputStream;
import java.io.Serializable;

/**
 * Created by IntelliJ IDEA.
 * User: andy
 * Date: Mar 24, 2010
 *
 * Extends XMLStreamReader with basic helper functions for parsing a wiseml with Stax.
 */
public class ExtendedXMLStreamReader implements Serializable{


    private XMLStreamReader reader;

    /**
     * Default Constructor.
     *
     * @param is the xml InputStream
     * @throws XMLStreamException if there is an xml syntax
     */
    public ExtendedXMLStreamReader(InputStream is) throws XMLStreamException {
        XMLInputFactory factory = XMLInputFactory.newInstance();
        factory.setProperty(XMLInputFactory.IS_NAMESPACE_AWARE, false);
        factory.setProperty(XMLInputFactory.IS_VALIDATING, false);
        factory.setProperty(XMLInputFactory.IS_COALESCING, false);
        reader = factory.createXMLStreamReader(is);
    }

    /**
     * @return the XMLStreamReader.
     */
    public XMLStreamReader getReader() {
        return reader;
    }

    public void close() throws XMLStreamException {
        reader.close();
    }
    /**
     * Reads a single position element. Starts at origin (or position) and ends at the following element.
     *
     * @return a Position object with the parsed information
     * @throws badSyntaxException throwed if a required element is missing or if a text doesn't match the expected type
     * @throws XMLStreamException is thrown if there is a problem with xml syntax.
     */
    protected Position position() throws badSyntaxException, XMLStreamException {

        float x, y, z = 0;
        double phi = 0, theta = 0;

        try{
            // Parse the x value - required
            x = Float.parseFloat(navigateRead("x"));

            // Parse the y value - required
            y = Float.parseFloat(navigateRead("y"));

            nextElement();

            // Parse the z value - optional
            if(check("z")){
                z = Float.parseFloat(elementText());
                nextElement();
            }

            // Parse the phi angle - optional
            if(check("phi")){                                         
                phi = Double.parseDouble(elementText());
                nextElement();
            }

            // Parse the theta angle - optional
            if (check("theta")){
                theta = Double.parseDouble(elementText());
                nextElement();
            }

        } catch(IllegalArgumentException e){
            throw new badSyntaxException("Illegal type in <posititon>");
        } catch (endDocumentException e) {
            throw new badSyntaxException("Unexpected end of document");
        }

        return new Position(x, y, z, phi, theta);

    }

    /**
     * Reads a single capability. The calling code has to move the stream to the next element.
     *
     * @return the new capability object.
     * @throws badSyntaxException is thrown if the element values don't have the expected format.
     * @throws XMLStreamException is thrown if there is a problem with xml syntax.
     */
    protected Capability capability() throws badSyntaxException, XMLStreamException{
        String name;
        Setup.Datatype dtype;
        Setup.Unit unit;
        String def;
        try {
            name = navigateRead("name");
            dtype = Setup.Datatype.valueOf(navigateRead("datatype"));
            unit = Setup.toUnit(navigateRead("unit"));
            def = navigateRead("default");

        } catch (IllegalArgumentException e) {
            throw new badSyntaxException("Illegal type in <capability>");
        } catch (endDocumentException e) {
            throw new badSyntaxException("Unexpected end of document");
        }
        return new Capability(name, dtype, unit, def);
    }

    /**
     * Reads a single Rssi object. The calling code has to move the stream to the next element.
     *
     * @return a new rssi object.
     * @throws badSyntaxException is thrown if the element values don't have the expected format.
     */
    protected Rssi rssi() throws badSyntaxException {
        Setup.Datatype dtype;
        Setup.Unit unit;
        String def;

        // An <rssi> should have 3 attributes
        // Otto: EHM... no? Only one?
        /*if (reader.getAttributeCount()!=3){
            throw new badSyntaxException("Wrong number of attributes in <rssi>");
        }*/
        try {
            // TODO dtype = Setup.Datatype.valueOf(attr("datatype"));
            // TODO unit = Setup.toUnit(attr("unit"));
            def = attr("value");

        } catch(IllegalArgumentException e){
            throw new badSyntaxException("Illegal type in <rssi>");
        }
        // return new Rssi(dtype, unit, def);
        return new Rssi(def);
    }

    /**
     * Navigates to an expected element and reads its text.
     *
     * @param expectedElement provide the name of the element to navigate to.
     * @return a String containing the parsed data or null.
     * @throws javax.xml.stream.XMLStreamException is thrown if there is a problem with xml syntax.
     * @throws walker.exceptions.badSyntaxException is thrown if after 20 elements no match is found.
     * @throws walker.exceptions.endDocumentException if the root element is closed.
     */
    public String navigateRead(String expectedElement) throws XMLStreamException, badSyntaxException, endDocumentException {
        // first navigate to the expected element,
        navigate(expectedElement);
        // move the reader to the element's text and return it.
        nextStaxElement();
        return text();
    }

    /**
     * Navigates to an expected element in the stream.
     *
     * @param expectedElement provide the name of the element to navigate to.
     * @throws XMLStreamException is thrown if there is no other element.
     * @throws badSyntaxException is thrown if after 200 elements no match is found.
     * @throws walker.exceptions.endDocumentException if the root element is closed.
     */
    public void navigate(String expectedElement) throws XMLStreamException, badSyntaxException, endDocumentException {
        int counter = 1;
        while(!check(expectedElement)){
            if ((counter > 300)){
                throw new badSyntaxException("expected element not found");
            }
            counter++;
            nextElement();
        }
    }

    /**
     * Safely navigates to the next element in the stream.
     *
     * @throws XMLStreamException is thrown if there is no other element.
     * @throws walker.exceptions.endDocumentException if the root element is closed.
     */
    public void nextElement() throws XMLStreamException, endDocumentException {
        nextStaxElement();
        while (reader.getEventType() != XMLStreamConstants.START_ELEMENT) {
            nextStaxElement();
        }
    }

    /**
     * Parses a single attribute of the current element. It is useful only if the current stream position is
     * in a XMLStreamConstants.START_ELEMENT. The stream position stays the same as to allow an elementText()
     * call, whick means that if no text is read, nextElement() should be called to move the stream to the
     * std position.
     *
     * @param attr the name of the attribute.
     * @return a String containing the attribute value.
     */
    public String attr(String attr) {
        if (reader.isStartElement() && reader.getAttributeCount()>0){
            return reader.getAttributeValue(null, attr);
        }
		return null;
    }

    /**
     * Parses the Text of the current wiseml element. It is useful only if the current stream position is
     * in a XMLStreamConstants.START_ELEMENT. In that case, it retrieves the current element's text, which in Stax
     * terms is the next element.
     *
     * @return a String containing the parsed data or null.
     * @throws XMLStreamException is thrown if there is a problem with xml syntax.
     * @throws badSyntaxException is thrown if there is no text.
     * @throws walker.exceptions.endDocumentException if the root element is closed.
     */
    public String elementText() throws XMLStreamException, badSyntaxException, endDocumentException {
        if (reader.isStartElement()){
            nextStaxElement();
            return text();
        }
		return null;
    }

    /**
     * Navigates to the next Stax element.
     * <p> Stax elements are:
     * <ul>
     * <li> XMLStreamConstants.START_ELEMENT
     * <li> XMLStreamConstants.END_ELEMENT
     * <li> XMLStreamConstants.PROCESSING_INSTRUCTION
     * <li> XMLStreamConstants.CHARACTERS
     * <li> XMLStreamConstants.COMMENT
     * <li> XMLStreamConstants.SPACE
     * <li> XMLStreamConstants.START_DOCUMENT
     * <li> XMLStreamConstants.END_DOCUMENT
     * <li> XMLStreamConstants.ENTITY_ REFERENCE
     * <li> XMLStreamConstants.ATTRIBUTE
     * <li> XMLStreamConstants.DTD
     * <li> XMLStreamConstants.CDATA
     * <li> XMLStreamConstants.NAMESPACE
     * <li> XMLStreamConstants.NOTATION_DECLARATION
     * <li> XMLStreamConstants.ENTITY_DECLARATION
     * </ul>
     * @throws XMLStreamException is thrown if there is no other element.
     * @throws walker.exceptions.endDocumentException if the root element is closed.
     */
    public void nextStaxElement() throws XMLStreamException, endDocumentException {

        if (reader.getEventType() == XMLStreamConstants.END_DOCUMENT){
            throw new endDocumentException();
        } else if(reader.hasNext()){
            reader.next();
        }
    }

    /**
     * Make sure that the current element is the one that is expected.
     *
     * @param expect the expected element name.
     * @return true if expected, else false.
     */
    public boolean check(String expect) {
        return (reader.hasName() && reader.isStartElement() && reader.getLocalName().equals(expect));
    }

    /**
     * Fail Safe retrieval of text. Text is retrieved only if stream is at a CDATA Stax element.
     *
     * @return a String containing the parsed data or null.
     * @throws badSyntaxException if no text is found.
     */
    public String text() throws badSyntaxException {
        String tmp;
        if (reader.hasText()) {
            tmp = reader.getText();
        } else {
            throw new badSyntaxException("there should be text here..!!");
        }
        return tmp;
    }

    
    public void debug(){

        if (reader.isStartElement()){
            print("<" + reader.getLocalName() + ">");
        } else if (reader.isEndElement()){
            print("</" + reader.getLocalName() + ">");
        } else if (reader.isCharacters()){
            print(reader.getText());
        } else if (reader.getEventType() == 7){
            print("DOC_START");
        } else {
            System.out.print(reader.getEventType());
        }

    }

    /**
     * Returns the current Location of the XMLStreamReader
     *
     * @return a String representation of the Location containing the characterOffset, the collumnNumber and
     * the lineNumber.
     */
    public String getLocation(){
        Location location = reader.getLocation();
        return "line#: " + location.getLineNumber() + "\ncharacter offset: " + location.getCharacterOffset();
    }


}
